# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Subcommand to find the principal components of a trajectory."""
import logging.config
import time
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional

import click
import holoviews as hv
import numpy as np
import pandas as pd
from holoviews import opts
from nptyping import Float
from nptyping import NDArray
from sklearn.decomposition import PCA

from .. import _MASK
from .. import create_logging_dict
from .. import PathLike
from ..libs.utils import get_positions
from ..libs.utils import reshape_positions

hv.extension("matplotlib")


@click.command("pca", short_help="Perform principal component analysis on a trajectory")
@click.option(
    "-s",
    "--top",
    "topology",
    metavar="FILE",
    default=Path.cwd().joinpath("input.top"),
    show_default=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Topology",
)
@click.option(
    "-f",
    "--traj",
    "trajectory",
    metavar="FILE",
    default=[
        Path.cwd().joinpath("input.nc"),
    ],
    show_default=True,
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Trajectory",
)
@click.option(
    "-o",
    "--outdir",
    metavar="DIR",
    default=Path.cwd(),
    show_default=True,
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help="Output directory",
)
@click.option(
    "-l",
    "--logfile",
    metavar="LOG",
    show_default=True,
    default=Path.cwd().joinpath("align_traj.log"),
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Log file",
)
@click.option(
    "--dt",
    "step",
    metavar="OFFSET",
    default=0,
    show_default=True,
    type=click.IntRange(min=0, clamp=True),
    help="Trajectory output offset (0 = single step)",
)
@click.option(
    "-m",
    "--mask",
    default="ca",
    show_default=True,
    type=click.Choice(_MASK.keys()),
    help="Atom selection",
)
@click.option(
    "-n",
    "--nmodes",
    "n_modes",
    metavar="NMODES",
    default=0,
    show_default=True,
    type=click.IntRange(min=0, clamp=True),
    help="Number of eigenmodes (0 = all components)",
)
@click.option("-w", "--whiten", is_flag=True, help="Whitens the data")
@click.option("--bias / --no-bias", help="Calculate with population bias (N vs N-1")
@click.option(
    "--dpi",
    default=600,
    show_default=True,
    type=click.IntRange(min=100, clamp=True),
    help="Resolution of the figure",
)
@click.option(
    "--it",
    "image_type",
    default="png",
    show_default=True,
    type=click.Choice("png pdf svg jpg".split(), case_sensitive=False),
    help="Image type",
)
@click.option("-v", "--verbose", is_flag=True, help="Noisy output")
def cli(
    topology: PathLike,
    trajectory: List[str],
    outdir: PathLike,
    logfile: PathLike,
    step: int,
    mask: str,
    n_modes: int,
    whiten: bool,
    bias: bool,
    dpi: int,
    image_type: str,
    verbose: bool,
) -> None:
    """Calculate principal components for a trajectory."""
    start_time: float = time.perf_counter()

    # Setup logging
    logging.config.dictConfig(create_logging_dict(logfile))
    logger: logging.Logger = logging.getLogger(__name__)

    out_dir = Path(outdir)
    step = step if step > 0 else 1

    # Extract positions and reshape to (n_frames, n_points * 3)
    logger.info("Loading trajectory positions")
    positions: NDArray[(Any, ...), Float] = get_positions(
        topology, trajectory, mask=_MASK[mask], stride=step
    )
    positions = reshape_positions(positions)
    n_samples, n_features = positions.shape
    n_components: Optional[int] = n_modes if n_modes > 0 else None

    logger.info("Calculating PCA")
    logger.warning(
        "Depending upon the size of the trajectory, this could take a while."
    )
    pca = PCA(n_components=n_components, svd_solver="full", whiten=whiten)

    projection = pd.DataFrame(pca.fit_transform(positions))
    projection.columns = [f"PC{_+1}" for _ in range(pca.n_components_)]
    projection.index.name = "Frame"

    logger.info("%d components were calculated.", pca.n_components_)
    explained_variance = pd.Series(pca.explained_variance_, name="Explained Variance")
    explained_variance.index += 1
    explained_variance.index.name = "Component"

    singular_values = pd.Series(pca.singular_values_, name="Singular Value")
    singular_values.index = explained_variance.index.copy()
    singular_values.index.name = explained_variance.index.name

    if bias and whiten:
        conversion = n_samples / (n_samples - 1)
        projection *= np.sqrt(conversion)
        explained_variance /= conversion

    ratio: pd.Series = explained_variance.cumsum() / explained_variance.sum() * 100.0
    ratio.name = "Percentage of Explained Variance"
    for percentage in range(80, 100, 5):
        logger.info(
            "%d components cover %.1f%% of the explained variance",
            ratio.where(ratio <= float(percentage)).dropna().size,
            percentage,
        )

    if pca.n_components_ >= 50:
        for component in np.fromiter([50, 100], dtype=int):
            logger.info(
                "%d components cover %.1f%% of the explained variance",
                component,
                ratio.iloc[component],
            )

    components = pd.DataFrame(pca.components_, index=projection.columns).T
    components.index += 1
    components.index.name = "Feature"

    # Save data
    data = dict(
        projection=projection,
        components=components,
        explained_variance=explained_variance,
        explained_variance_ratio=ratio,
        singular=singular_values,
    )
    for key, value in data.items():
        with out_dir.joinpath(f"{key}.csv").open(mode="w") as w:
            logger.info("Saving %s to %s", key, w.name)
            value.reset_index().to_csv(w, float_format="%.6f", index=False)
        with out_dir.joinpath(f"{key}.npy").open(mode="wb") as w:  # type: ignore
            logger.info("Saving %s to %s", key, w.name)
            np.save(w, value.values)

    # Plot explained variance ratio
    filename = out_dir.joinpath("explained_variance_ratio").with_suffix(
        "." + image_type.lower()
    )
    logger.info("Saving explained variance ratio to %s", filename)
    curve = hv.Curve(ratio, "Component", "Percentage of Explained Variance")
    points = hv.Points(ratio, ["Component", "Percentage of Explained Variance"])
    overlay = curve * points
    overlay.opts(
        opts.Curve(linewidth=1.5, color="purple"),
        opts.Points(s=1.5, marker=".", color="yellow"),
    )
    hv.output(dpi=dpi)
    hv.save(
        overlay,
        filename=filename,
        backend="matplotlib",
        title="Explained Variance Percentage",
    )

    stop_time: float = time.perf_counter()
    dt: float = stop_time - start_time
    struct_time: time.struct_time = time.gmtime(dt)
    if verbose:
        output: str = time.strftime("%H:%M:%S", struct_time)
        logger.info(f"Total execution time: {output}")



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Draw and save figures for QAA."""
import itertools
from pathlib import Path
from typing import Optional

import holoviews as hv
import pandas as pd
from colorcet import glasbey_cool
from colorcet import glasbey_light

from .. import PathLike

hv.extension("matplotlib")

_empty_dataframe = pd.DataFrame()


class Figure:
    """Create a plot of 2D and 3D plots."""

    def __init__(
        self,
        *,
        azim: int = 120,
        elevation: int = 30,
    ) -> None:
        """Visualize data via a graphical image.

        Parameters
        ----------
        azim : int
            Azimuth rotation for 3D plot
        elevation : int
            Elevation of 3D plot
        """
        self._figure: Optional[hv.Layout] = None
        self._azimuth: int = azim
        self._elevation: int = elevation

    @property
    def figure(self) -> hv.Layout:
        """Return a holoviews Layout.

        Returns
        -------
        hv.Layout
            A multicolumn 2D and 3D scatter plots
        """
        return self._figure

    @property
    def azimuth(self) -> int:
        """Return the azimuth of the 3D scatter plot.

        Returns
        -------
        int
            Azimuth
        """
        return self._azimuth

    @azimuth.setter
    def azimuth(self, azimuth: int) -> None:
        """Set azimuth of the 3D scatter plot.

        Parameters
        ----------
        azimuth : int
            Azimuth
        """
        self._azimuth = azimuth

    @property
    def elevation(self) -> int:
        """Set elevation of the 3D scatter plot.

        Returns
        -------
        int
            Elevation of the 3D plot
        """
        return self._elevation

    @elevation.setter
    def elevation(self, elevation: int) -> None:
        """Set elevation of the 3D scatter plot.

        Parameters
        ----------
        elevation : int
            Elevation
        """
        self._elevation = elevation

    def draw(
        self,
        data: pd.DataFrame,
        *,
        centers: pd.DataFrame = _empty_dataframe,
    ) -> None:
        """Draw the first three components in subplots.

        Parameters
        ----------
        data : pd.DataFrame
            Matrix with shape (n_samples, n_components)
        centers : pd.DataFrame, optional
            Vector of cluster centers (n_components, )

        Notes
        -----
        Four subplots are created.
        * component 1 vs. component 2
        * component 1 vs. component 3
        * component 2 vs. component 3
        * 3D plot
        """
        table = hv.Table(data)
        kdims = table.data.columns[2:5].to_list()
        vdims = table.data.columns[:2].to_list()
        centroid = hv.Table(centers) if not centers.empty else hv.Table([])

        points = [
            hv.Points(table, [i, j], vdims=vdims).opts(
                show_legend=False,
                marker=".",
                s=5,
                cmap=glasbey_cool,
                color_index="Cluster",
            )
            for i, j in itertools.combinations(kdims, 2)
        ]
        if centroid:
            scatter = [
                hv.Points(centroid, kdims=[i, j], vdims="Cluster").opts(
                    show_legend=False,
                    marker=".",
                    s=15,
                    cmap=glasbey_light,
                    color_index="Cluster",
                )
                for i, j in itertools.combinations(kdims, 2)
            ]
            points = [i * j for i, j in zip(points, scatter)]
        self._figure = hv.Layout(points)

        scatter3d = hv.Scatter3D(data, kdims, vdims=vdims).opts(
            show_legend=False,
            marker=".",
            s=5,
            cmap=glasbey_cool,
            color_index="Cluster",
        )
        if centroid:
            scatter3d *= hv.Scatter3D(centers, kdims=kdims, vdims="Cluster").opts(
                show_legend=False,
                marker=".",
                s=25,
                cmap=glasbey_light,
                color_index="Cluster",
            )
        self._figure += scatter3d
        self._figure.cols(2)

    def save(
        self, filename: PathLike, *, dpi: int = 600, title: Optional[str] = None
    ) -> None:
        """Save the image to disk.

        Parameters
        ----------
        filename : PathLike
            Image file
        dpi : int
            Image resolution
        title : str
            Figure title
        """
        hv.output(dpi=dpi)
        hv.save(
            self._figure,
            filename=filename,
            fmt=Path(filename).suffix[1:],
            backend="matplotlib",
            title=title,
        )



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Align a trajectory using the Kabsch method.

Notes
-----
Kabsch functions originally from
https://github.com/charnley/rmsd/blob/master/rmsd/calculate_rmsd.py
"""
from __future__ import annotations

import logging
from typing import Any
from typing import Optional
from typing import Tuple

from nptyping import Float
from nptyping import NDArray
from scipy import linalg
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

from .utils import rmse

logger: logging.Logger = logging.getLogger(__name__)


class Kabsch(TransformerMixin, BaseEstimator):
    """Kabsch alignment method."""

    def __init__(self, verbose: bool = True) -> None:
        """Align one array to another using the Kabsch method.

        Parameters
        ----------
        verbose : bool
            Print data to the log
        """
        self.rotation_: NDArray[(Any, ...), Float]
        self.translation_: NDArray[(Any, ...), Float]
        self.error: float
        self.verbose = verbose

    def fit(
        self, mobile: NDArray[(Any, ...), Float], reference: NDArray[(Any, ...), Float]
    ) -> Kabsch:
        """Center and fit `mobile` onto `reference` using the Kabsch method.

        Parameters
        ----------
        mobile, reference : array_like
            Arrays with shape (n_points, 3)

        Returns
        -------
        self
        """
        mobile = check_array(mobile, copy=self.verbose)
        reference = check_array(reference, copy=self.verbose)
        new_mobile, _ = self._center(mobile)
        new_ref, ref_mean = self._center(reference)

        self.rotation_ = self._fit(new_mobile, new_ref)
        self.translation_ = ref_mean.copy()

        return self

    def transform(
        self,
        mobile: NDArray[(Any, ...), Float],
        reference: Optional[NDArray[(Any, ...), Float]] = None,
    ) -> NDArray[(Any, ...), Float]:
        """Align `mobile` to `reference`.

        Parameters
        ----------
        mobile, reference : array_like
            Arrays with shape (n_points, 3)

        Returns
        -------
        array_like
            Array with shape (n_points, 3)
        """
        check_is_fitted(self)

        mobile = StandardScaler(with_std=False).fit_transform(mobile)
        new_mobile: NDArray[(Any, ...), Float] = (
            mobile @ self.rotation_ + self.translation_
        )
        self.error = rmse(mobile, new_mobile)
        return new_mobile

    def _center(
        self, positions: NDArray[(Any, ...), Float]
    ) -> Tuple[NDArray[(Any, ...), Float], NDArray[(Any, ...), Float]]:
        scale = StandardScaler(with_std=False)
        new_positions: NDArray[(Any, ...), Float] = scale.fit_transform(positions)
        return new_positions, scale.mean_

    def _fit(
        self, mobile: NDArray[(Any, ...), Float], reference: NDArray[(Any, ...), Float]
    ) -> NDArray[(Any, ...), Float]:
        """Align `mobile` to `reference` using the Kabsch algorithm.

        For more info see https://en.wikipedia.org/wiki/Kabsch_algorithm

        Parameters
        ----------
        mobile, reference : NDArray[(Any, ...), Float]
            Arrays with shape (n_points, 3)

        Returns
        -------
        ArrayType
            Rotation matrix with shape (3, 3)

        Notes
        -----
        Using the Kabsch algorithm with two sets of paired point `arr1` and `arr2`, centered
        around the centroid. Each vector set is represented with the shape  (n_points,
        n_dims) where n_dims is the dimension of the space.  The algorithm works in three
        steps:

        * a centroid translation of `arr1` and `arr2` (completed before this function call)
        * the computation of a covariance matrix `covar`
        * computation of the optimal rotation matrix `rot_matrix`
        """
        # Computation of the covariance matrix
        covar: NDArray[(Any, ...), Float] = mobile.T @ reference

        # Computation of the optimal rotation matrix
        # This can be done using singular value decomposition (SVD)
        # Getting the sign of the det(V)*(W) to decide
        # whether we need to correct our rotation matrix to ensure a
        # right-handed coordinate system.
        # And finally calculating the optimal rotation matrix U
        # see http://en.wikipedia.org/wiki/Kabsch_algorithm
        u, s, vh = linalg.svd(covar)
        if (linalg.det(u) * linalg.det(vh)) < 0.0:
            s[-1] = -s[-1]
            u[:, -1] = -u[:, -1]

        return u @ vh



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Subcommand to plot data."""
import logging.config
import time
from pathlib import Path
from typing import Tuple

import click
import numpy as np
import pandas as pd

from .. import create_logging_dict
from .. import PathLike
from ..libs.figure import Figure


@click.command("cluster", short_help="Plot data from QAA.")
@click.option(
    "-i",
    "--infile",
    metavar="FILE",
    default=Path.cwd().joinpath("input.csv"),
    show_default=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Data file for analysis",
)
@click.option(
    "--label",
    metavar="FILE",
    default=Path.cwd().joinpath("labels.npy"),
    show_default=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Cluster labels",
)
@click.option(
    "-c",
    "--centroid",
    metavar="FILE",
    default=Path.cwd().joinpath("centroids.csv"),
    show_default=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Cluster labels",
)
@click.option(
    "-o",
    "--outfile",
    metavar="FILE",
    default=Path.cwd().joinpath("cluster.png"),
    show_default=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Image file",
)
@click.option(
    "-l",
    "--logfile",
    metavar="LOG",
    show_default=True,
    default=Path.cwd() / "plot.log",
    type=click.Path(exists=False, file_okay=True, resolve_path=True),
    help="Log file",
)
@click.option(
    "--axes",
    metavar="AXES",
    nargs=3,
    default=(0, 1, 2),
    type=click.IntRange(min=0, clamp=True),
    help="Components to plot",
)
@click.option("--ica / --pca", "method", default=True, help="Type of data")
@click.option(
    "--dpi",
    metavar="DPI",
    default=600,
    show_default=True,
    type=click.IntRange(min=100, clamp=True),
    help="Resolution of the figure",
)
@click.option(
    "--azim",
    "azimuth",
    metavar="AZIMUTH",
    default=120,
    show_default=True,
    type=click.IntRange(min=0, max=359, clamp=True),
    help="Azimuth rotation of 3D plot",
)
@click.option(
    "--elev",
    "elevation",
    metavar="ELEVATION",
    default=30,
    show_default=True,
    type=click.IntRange(min=0, max=90, clamp=True),
    help="Elevation of 3D plot",
)
@click.option("--cluster", is_flag=True, help="Cluster analysis")
@click.option("-v", "--verbose", is_flag=True, help="Noisy output")
def cli(
    infile: PathLike,
    label: PathLike,
    centroid: PathLike,
    outfile: PathLike,
    logfile: PathLike,
    axes: Tuple[int, int, int],
    method: bool,
    dpi: int,
    azimuth: int,
    elevation: int,
    cluster: bool,
    verbose: bool,
) -> None:
    """Visualize the data."""
    start_time: float = time.perf_counter()
    in_file = Path(infile)

    # Setup logging
    logging.config.dictConfig(create_logging_dict(logfile))
    logger: logging.Logger = logging.getLogger(__name__)

    data_method = "ica" if method else "pca"
    sorted_axes = np.sort(axes)
    features = [f"{data_method[:2].upper()}{_+1:d}" for _ in sorted_axes]

    # Load data
    logger.info("Loading %s", in_file)

    index = "Frame"
    data = read_file(in_file, index=index)
    if data.empty:
        raise SystemExit(f"Unable to read {in_file}")

    data.columns = (
        [f"{data_method[:2].upper()}{_+1:d}" for _ in range(data.columns.size)]
        if np.issubdtype(data.columns, int)
        else data.columns
    )
    try:
        data = pd.concat([data["Cluster"], data[features].reset_index()], axis=1)
    except KeyError:
        data = data[features].reset_index()

    # Load labels, if exists
    centroid_data: pd.DataFrame = pd.DataFrame()
    if cluster:
        label_data = read_file(Path(label))
        if "Cluster" not in data.columns and not label_data.empty:
            label_data.columns = ["Cluster"]
            data = pd.concat([label_data, data], axis=1)

        # Load centroid data, if exists
        centroid_data = read_file(Path(centroid), index="Cluster")
        if not centroid_data.empty:
            centroid_data = centroid_data.set_index("Cluster")
            centroid_data.columns = features
            centroid_data = centroid_data.reset_index()
    else:
        n_samples, _ = data.shape
        label_data = pd.Series(np.zeros(n_samples, dtype=int), name="Cluster")
        if "Cluster" not in data.columns:
            data = pd.concat([label_data, data], axis=1)
        else:
            data["Cluster"] = label_data.copy()

    # Prepare cluster analysis
    figure = Figure(azim=azimuth, elevation=elevation)
    logger.info("Preparing figures")
    figure.draw(data, centers=centroid_data)

    logger.info("Saving visualization data to %s", outfile)
    figure.save(outfile, dpi=dpi)

    stop_time: float = time.perf_counter()
    dt: float = stop_time - start_time
    struct_time: time.struct_time = time.gmtime(dt)
    if verbose:
        output: str = time.strftime("%H:%M:%S", struct_time)
        logger.info(f"Total execution time: {output}")


def read_file(filename: Path, *, index: str = "") -> pd.DataFrame:
    """Read a file and return a DataFrame.

    Parameters
    ----------
    filename : Path
        File to open
    index : str
        Index name

    Returns
    -------
    pd.DataFrame
        Parsed data
    """
    try:
        df = (
            pd.read_csv(filename, header=0)
            if filename.suffix == ".csv"
            else pd.DataFrame(np.load(filename))
        )
        if not df.index.name:
            df.index.name = index
    except Exception:
        df = pd.DataFrame()
    return df



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Various utilities."""
import glob
from typing import Any
from typing import List
from typing import Optional

import mdtraj as md
import numpy as np
from mdtraj.utils import in_units_of
from nptyping import Float
from nptyping import NDArray

from .. import PathLike


def get_average_structure(
    topology: PathLike,
    trajectory: List[str],
    *,
    mask: str = "all",
    stride: Optional[int] = None,
) -> md.Trajectory:
    """Compute the average structure of a trajectory.

    Parameters
    ----------
    topology : PathLike
        Topology file
    trajectory : list of str
        List of trajectory files
    mask : str
        Atom selection
    stride : int, optional
        Number of steps to read

    Returns
    -------
    Trajectory
        The average positions
    """
    n_frames: int = 0
    positions_: List[NDArray[(Any, ...), Float]] = []
    indices: Optional[NDArray[(Any, ...), Float]] = (
        md.load_topology(topology).select(mask) if mask != "all" else None
    )
    filenames = (
        glob.iglob(*trajectory)
        if len(trajectory) == 1 and "*" in "".join(trajectory)
        else trajectory
    )

    for filename in filenames:
        for frames in md.iterload(
            filename, top=topology, atom_indices=indices, stride=stride
        ):
            n_frames += frames.n_frames
            coordinates = frames.xyz.sum(axis=0)
            positions_.append(coordinates)

    # MDTraj stores positions in nanometers; we convert it to Ångstroms.
    positions: NDArray[(Any, ...), Float] = np.asfarray(positions_)
    frames.xyz = positions.sum(axis=0) / n_frames
    frames.unitcell_angles = frames.unitcell_angles[0, :]
    frames.unitcell_lengths = frames.unitcell_lengths[0, :]
    return frames


def get_positions(
    topology: PathLike,
    trajectory: List[str],
    *,
    mask: str = "all",
    stride: Optional[int] = None,
) -> NDArray[(Any, ...), Float]:
    """Read a molecular dynamics trajectory and retrieve the coordinates.

    Parameters
    ----------
    topology : PathLike
        Topology file
    trajectory : list of str
        Trajectory file
    mask : str
        Selection criterion for coordinates
    stride : int, optional
        Number of steps to read

    Returns
    -------
    NDArray
        The coordinates with shape (n_frames / step, n_atoms, 3)
    """
    top: md.Topology = md.load_topology(topology)
    selection: Optional[NDArray[(Any, ...), Float]] = (
        top.select(mask) if mask != "all" else None
    )
    filenames = (
        glob.iglob(*trajectory)
        if len(trajectory) == 1 and "*" in "".join(trajectory)
        else trajectory
    )

    # MDTraj stores positions in nanometers; we convert it to Ångstroms.
    positions: NDArray[(Any, ...), Float] = np.concatenate(
        [
            frames.xyz
            for filename in filenames
            for frames in md.iterload(
                filename, top=top, atom_indices=selection, stride=stride
            )
        ],
        axis=0,
    )
    if not (
        ".gro" in "".join(filenames)
        or ".xtc" in "".join(filenames)
        or ".trj" in "".join(filenames)
        or ".tng" in "".join(filenames)
    ):
        in_units_of(positions, "nanometer", "angstroms", inplace=True)
    return positions


def reshape_positions(
    positions: NDArray[(Any, ...), Float]
) -> NDArray[(Any, ...), Float]:
    """Reshape a n * m * 3 trajectory to a n * (m * 3) 2D matrix.

    Parameters
    ----------
    positions : array_like
        A 3-D matrix with shape (n_frames, n_atoms, 3)

    Returns
    -------
    NDArray
        A 2-D array with shape (n_frames, n_atoms * 3)
    """
    n_frames, n_atoms, n_dims = positions.shape
    return positions.reshape((n_frames, n_atoms * n_dims))


def rmse(
    mobile: NDArray[(Any, ...), Float], reference: NDArray[(Any, ...), Float]
) -> float:
    """Calculate the root-mean-square error between two arrays.

    Parameters
    ----------
    mobile : array_like
        coordinates
    reference : array_like
        coordinates

    Returns
    -------
    float
        The error difference between the two arrays
    """
    return float(np.linalg.norm(mobile - reference))



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Cluster data into regions."""
import glob
import logging.config
import time
from pathlib import Path
from typing import Any
from typing import Sequence
from typing import Tuple

import click
import mdtraj as md
import numpy as np
import pandas as pd
from nptyping import Float
from nptyping import NDArray
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture

from .. import create_logging_dict
from .. import PathLike


@click.command("cluster", short_help="Plot data from QAA.")
@click.option(
    "-s",
    "--top",
    "topology",
    metavar="FILE",
    default=Path.cwd().joinpath("input.top"),
    show_default=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Topology",
)
@click.option(
    "-f",
    "--traj",
    "trajectory",
    metavar="FILE",
    default=[
        Path.cwd().joinpath("input.nc"),
    ],
    show_default=True,
    multiple=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Trajectory",
)
@click.option(
    "-i",
    "--infile",
    metavar="FILE",
    default=Path.cwd().joinpath("input.csv"),
    show_default=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Data file for analysis",
)
@click.option(
    "-o",
    "--outdir",
    metavar="DIR",
    default=Path.cwd(),
    show_default=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
    help="Clustered image file",
)
@click.option(
    "-l",
    "--logfile",
    metavar="LOG",
    show_default=True,
    default=Path.cwd() / "image.log",
    type=click.Path(exists=False, file_okay=True, resolve_path=True),
    help="Log file",
)
@click.option(
    "--axes",
    metavar="AXES",
    nargs=3,
    default=(0, 1, 2),
    show_default=True,
    type=click.IntRange(min=0, clamp=True),
    help="Components",
)
@click.option("--ica / --pca", "method", default=True, help="Type of data")
@click.option("--gmm / --kmeans", "cluster", default=True, help="Clustering method")
@click.option(
    "--iter",
    "max_iter",
    metavar="MAXITER",
    default=200,
    show_default=True,
    type=click.IntRange(min=1, clamp=True),
    help="Maximum number of iterations for clustering",
)
@click.option(
    "--tol",
    metavar="TOL",
    default=0.001,
    show_default=True,
    type=click.FloatRange(min=0.0, max=1.0, clamp=True),
    help="Maximum tolerance for clustering",
)
@click.option(
    "-n",
    "--nclusters",
    "n_clusters",
    default=3,
    show_default=True,
    type=click.IntRange(min=2, max=4, clamp=True),
    help="Number of cluster centers to plot",
)
@click.option(
    "--dp",
    "n_points",
    default=1,
    show_default=True,
    type=click.IntRange(min=1, clamp=True),
    help="Number of points to skip for plotting",
)
@click.option(
    "--save", is_flag=True, help="Save structures from corresponding cluster center"
)
@click.option("-v", "--verbose", is_flag=True, help="Noisy output")
def cli(
    topology: PathLike,
    trajectory: Sequence[str],
    infile: PathLike,
    outdir: PathLike,
    logfile: PathLike,
    axes: Tuple[int, int, int],
    method: bool,
    cluster: bool,
    max_iter: int,
    tol: float,
    n_clusters: int,
    n_points: int,
    save: bool,
    verbose: bool,
) -> None:
    """Perform cluster analysis on the provided data."""
    start_time: float = time.perf_counter()
    in_file = Path(infile)
    out_dir = Path(outdir)

    # Setup logging
    logging.config.dictConfig(create_logging_dict(logfile))
    logger: logging.Logger = logging.getLogger(__name__)

    data_method = "ica" if method else "pca"
    sorted_axes = np.sort(axes)
    features = [f"{data_method[:2].upper()}{_+1:d}" for _ in sorted_axes]

    # Load data
    data: pd.DataFrame
    try:
        try:
            data = pd.DataFrame(np.load(in_file)[:, axes], columns=features)
            data.index.name = "Frame"
        except ValueError:
            data = pd.read_csv(in_file, header=0, index_col=0)[features]
    except BaseException:
        raise

    # Select clustering method and cluster data
    clustering = (
        GaussianMixture(
            n_components=n_clusters,
            max_iter=max_iter,
            tol=tol,
        )
        if cluster
        else KMeans(
            n_clusters=n_clusters,
            max_iter=max_iter,
            tol=tol,
        )
    )
    labels = pd.Series(clustering.fit_predict(data), name="Cluster")
    centroids = pd.DataFrame(
        clustering.means_ if cluster else clustering.cluster_centers_, columns=features
    )
    centroids.index.name = "Cluster"

    data = pd.concat([labels, data.reset_index()], axis=1)

    # Prepare dataframe
    with out_dir.joinpath(f"{data_method}-cluster.csv").open(mode="w") as w:
        logger.info("Saving cluster data to %s", w.name)
        data.to_csv(w, index=False, float_format="%.6f")

    with out_dir.joinpath(f"{data_method}-cluster.npy").open(mode="wb") as wb:
        logger.info("Saving cluster data to %s", wb.name)
        np.save(wb, data.drop(["Cluster", "Frame"], axis=1))
    with out_dir.joinpath(f"{data_method}-labels.npy").open("wb") as wb:
        logger.info("Saving label data to %s", wb.name)
        np.save(wb, data["Cluster"])

    with out_dir.joinpath(f"{data_method}-centroids.csv").open("w") as w:
        logger.info("Saving centroids to %s", w.name)
        centroids.reset_index().to_csv(w, float_format="%.6f", index=False)
    with out_dir.joinpath(f"{data_method}-centroids.npy").open("wb") as wb:
        logger.info("Saving centroids to %s", wb.name)
        np.save(wb, centroids)

    if save:
        # Find all trajectories and determine total frames per trajectory
        filenames: Sequence[str] = glob.glob(*trajectory)
        frames: NDArray[(Any, ...), Float] = np.asarray(
            [
                sum([_.n_frames for _ in md.iterload(filename, top=topology)])
                for filename in filenames
            ],
            dtype=int,
        ).cumsum()

        for i, center in enumerate(centroids.iterrows()):
            idx: int = find_closest_point(center[1], data[features])
            file_no: int = int(np.searchsorted(frames, idx))
            traj: md.Trajectory = md.load_frame(filenames[file_no], idx, top=topology)

            filename = out_dir.joinpath(f"{data_method}-cluster{i}_frame{idx}.pdb")
            logger.info("Saving frame %d of cluster %d to %s", idx, i, filename)
            traj.save(filename.as_posix())

    stop_time: float = time.perf_counter()
    dt: float = stop_time - start_time
    struct_time: time.struct_time = time.gmtime(dt)
    if verbose:
        output: str = time.strftime("%H:%M:%S", struct_time)
        logger.info(f"Total execution time: {output}")


def find_closest_point(
    point: NDArray[(Any, ...), Float], data: NDArray[(Any, ...), Float]
) -> int:
    """Locate a point in the `data` closest to the `point`.

    Parameters
    ----------
    point : NDArray[(Any, ...), Float]
        Point of interest with size (n_features, )
    data : NDArray[(Any, ...), Float]
        Data to search with shape (n_samples, n_features)

    Returns
    -------
    int
        Index of value closes to `point`
    """
    distance: NDArray[(Any, ...), Float] = np.linalg.norm(data - point, axis=1)
    return int(np.where(distance == distance.min())[0])



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020—2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mqaa` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``qaa.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``qaa.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import logging
import sys
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import click
from click import core

from . import __version__

CONTEXT_SETTINGS = dict(
    auto_envvar_prefix="COMPLEX", help_option_names=["-h", "--help"]
)
logger = logging.getLogger()


class _Context(click.Context):
    """Context manager for click command-line interface."""

    def __init__(self) -> None:
        self.verbose = False
        self.home = Path.home()

    def log(self, fmt: str, *messages: Tuple[str]) -> None:
        """Log a message to stderr."""
        if messages:
            fmt %= messages
        click.echo(fmt, file=sys.stderr)

    def vlog(self, fmt: str, *messages: Tuple[str]) -> None:
        """Log a message to stderr only if verbose is enabled."""
        if self.verbose:
            self.log(fmt, *messages)


pass_context = click.make_pass_decorator(_Context, ensure=True)
cmd_folder = Path(__file__).parent.joinpath("commands").resolve()


class _ComplexCLI(click.MultiCommand):
    """Complex command-line options with subcommands for qaa."""

    def list_commands(self, ctx: click.Context) -> List[str]:
        """List available commands.

        Parameters
        ----------
        ctx : Context
            click context

        Returns
        -------
        List of str
            List of available commands
        """
        commands: List[str] = []
        for filename in cmd_folder.iterdir():
            if filename.suffix == ".py" and filename.stem.startswith("cmd_"):
                commands.append(filename.stem[4:])
        commands.sort()
        return commands

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[core.Command]:
        """Run the selected command.

        Parameters
        ----------
        ctx : Context
            click context
        cmd_name : str
            command name

        Returns
        -------
            The chosen command if present
        """
        try:
            if sys.version_info[0] == 2:
                cmd_name = cmd_name.encode("ascii", "replace")
            mod = __import__("qaa.commands.cmd_" + cmd_name, None, None, ["cli"])
        except ImportError:
            return None
        return mod.cli


@click.command(cls=_ComplexCLI, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__)
def main() -> None:
    """Run main command-line interface."""
    pass



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Quasi-Anharmonic Analysis."""
import logging
import os
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING
from typing import TypeVar

logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if TYPE_CHECKING:
    PathLike = TypeVar("PathLike", str, os.PathLike[str])
else:
    PathLike = TypeVar("PathLike", str, os.PathLike)

__version__: str = "1.0.0-rc1"

_MASK: Dict[str, str] = dict(
    ca="protein and name CA",
    cab="protein and name =~ 'C[AB]'",
    back="protein and backbone",
    side="protein and not backbone and not (element =~ 'H')",
    noh="protein and not (element =~ 'H')",
    all="all",
)


def create_logging_dict(logfile: PathLike) -> Dict[str, Any]:
    """Configure the logger.

    Parameters
    ----------
    logfile : PathLike
        Filename for log output.

    Returns
    -------
    Dict
        Configuration data for logging.

    Raises
    ------
    ValueError
        If a filename is not defined.
    """
    if not logfile:
        raise ValueError("Filename not defined.")

    logger_dict = dict(
        version=1,
        disable_existing_loggers=False,  # this fixes the problem
        formatters=dict(
            standard={
                "class": "logging.Formatter",
                "format": "%(name)-12s %(levelname)-8s %(message)s",
            },
            detailed={
                "class": "logging.Formatter",
                "format": ("%(asctime)s %(name)-15s %(levelname)-8s " "%(message)s"),
                "datefmt": "%m-%d-%y %H:%M",
            },
        ),
        handlers=dict(
            console={
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "standard",
            },
            file={
                "class": "logging.FileHandler",
                "filename": logfile,
                "level": "INFO",
                "mode": "w",
                "formatter": "detailed",
            },
        ),
        root=dict(level="INFO", handlers=["console", "file"]),
    )
    return logger_dict



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""CLI to compute quasi-anharmonic analysis."""
import logging.config
import time
from pathlib import Path
from typing import Any
from typing import List
from typing import Union

import click
import numpy as np
import pandas as pd
from nptyping import Float
from nptyping import NDArray
from sklearn.decomposition import FastICA

from .. import _MASK
from .. import create_logging_dict
from .. import PathLike
from ..decomposition.jade import JadeICA
from ..libs.utils import get_positions
from ..libs.utils import reshape_positions


@click.command("qaa", short_help="Perform quasi-anharmonic analysis of a trajectory")
@click.option(
    "-s",
    "--top",
    "topology",
    metavar="FILE",
    default=Path.cwd().joinpath("input.top"),
    show_default=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Topology",
)
@click.option(
    "-f",
    "--traj",
    "trajectory",
    metavar="FILE",
    default=[
        Path.cwd().joinpath("input.nc"),
    ],
    show_default=True,
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Trajectory",
)
@click.option(
    "-o",
    "--outdir",
    metavar="DIR",
    default=Path.cwd(),
    show_default=True,
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help="Output directory",
)
@click.option(
    "-l",
    "--logfile",
    metavar="LOG",
    show_default=True,
    default=Path.cwd().joinpath("align_traj.log"),
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Log file",
)
@click.option(
    "--dt",
    "step",
    metavar="OFFSET",
    default=0,
    show_default=True,
    type=click.IntRange(min=0, clamp=True),
    help="Trajectory output offset (0 = single step)",
)
@click.option(
    "-m",
    "--mask",
    default="ca",
    show_default=True,
    type=click.Choice(_MASK.keys()),
    help="Atom selection",
)
@click.option(
    "-n",
    "--nmodes",
    "n_modes",
    metavar="NMODES",
    default=0,
    show_default=True,
    type=click.IntRange(min=0, clamp=True),
    help="Number of eigenmodes (0 = all components)",
)
@click.option("--jade / --fastica", "method", default=True, help="QAA method")
@click.option("-w", "--whiten", is_flag=True, help="Whitens the data")
@click.option(
    "--iter",
    "max_iter",
    metavar="MAXITER",
    default=200,
    show_default=True,
    type=click.IntRange(min=1, clamp=True),
    help="Maximum number of iterations for FastICA",
)
@click.option(
    "--tol",
    metavar="TOL",
    default=0.001,
    show_default=True,
    type=click.FloatRange(min=0.0, max=1.0, clamp=True),
    help="Maximum number of iterations for FastICA",
)
@click.option("-v", "--verbose", is_flag=True, help="Noisy output")
def cli(
    topology: PathLike,
    trajectory: List[str],
    outdir: PathLike,
    logfile: PathLike,
    step: int,
    mask: str,
    n_modes: int,
    method: bool,
    whiten: bool,
    max_iter: int,
    tol: float,
    verbose: bool,
) -> None:
    """Perform quasi-anharmonic analysis on a trajectory."""
    start_time: float = time.perf_counter()

    # Setup logging
    logging.config.dictConfig(create_logging_dict(logfile))
    logger: logging.Logger = logging.getLogger(__name__)

    out_dir = Path(outdir)
    step = step if step > 0 else 1

    # Extract positions and reshape to (n_frames, n_points * 3)
    logger.info("Loading trajectory positions")
    positions: NDArray[(Any, ...), Float] = get_positions(
        topology, trajectory, mask=_MASK[mask], stride=step
    )
    positions = reshape_positions(positions)
    n_samples, n_features = positions.shape
    n_components: int = n_modes if n_modes > 0 else min(n_samples, n_features)

    logger.info("Calculating PCA")
    logger.warning(
        "Depending upon the size of the trajectory, this could take a while."
    )

    qaa_method = "Jade" if method else "Fast"
    logger.info("Running QAA using %sICA", qaa_method)
    logger.warning(
        "Depending upon the size of your system and number of modes, "
        "this could take a while..."
    )
    qaa: Union[JadeICA, FastICA] = (
        JadeICA(n_components=n_components)
        if method
        else FastICA(
            n_components=n_components,
            whiten=whiten,
            fun="cube",
            max_iter=max_iter,
            tol=tol,
            random_state=0,
        )
    )
    signals = pd.DataFrame(qaa.fit_transform(positions))
    signals.columns = [f"IC{_+1}" for _ in range(signals.columns.size)]
    signals.index.name = "Frame"

    # Save unmixed signals
    with out_dir.joinpath("qaa-signals.csv").open(mode="w") as w:
        logger.info("Saving QAA data to %s", w.name)
        signals.reset_index().to_csv(w, float_format="%.6f", index=False)
    with out_dir.joinpath("qaa-signals.npy").open(mode="wb") as w:  # type: ignore
        logger.info("Saving QAA data to %s", w.name)
        np.save(w, signals)

    # Save unmixing matrix
    unmixing = pd.DataFrame(qaa.components_, index=signals.columns)
    unmixing.columns += 1
    with out_dir.joinpath("unmixing_matrix.csv").open(mode="w") as w:
        logger.info("Saving QAA unmixing matrix to %s", w.name)
        unmixing.reset_index().to_csv(w, float_format="%.6f", index=False)
    with out_dir.joinpath("unmixing_matrix.npy").open(mode="wb") as w:  # type: ignore
        logger.info("Saving QAA unmixing matrix to %s", w.name)
        np.save(w, qaa.components_)

    stop_time: float = time.perf_counter()
    dt: float = stop_time - start_time
    struct_time: time.struct_time = time.gmtime(dt)
    if verbose:
        output: str = time.strftime("%H:%M:%S", struct_time)
        logger.info(f"Total execution time: {output}")



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Functions to align coordinate files."""
import logging
from typing import Any

import numpy as np
from nptyping import Float
from nptyping import NDArray

from .kabsch import Kabsch
from .utils import rmse

logger = logging.getLogger(__name__)


def align_trajectory(
    mobile: NDArray[(Any, ...), Float],
    reference: NDArray[(Any, ...), Float],
    *,
    tol: float = 1e-6,
    verbose: bool = True,
) -> NDArray[(Any, ...), Float]:
    """Align `mobile` to `reference` using the Kabsch method.

    Parameters
    ----------
    mobile : array_like
        Positions of shape (n_frames, n_points, n_dims)
    reference : array_like
        Positions of shape (n_points, n_dims)
    tol : float
        Tolerance level for alignment
    verbose : bool
        Print alignment progress

    Returns
    -------
    array_like
        Aligned mobile array of shape (n_frames, n_points, n_dims)
    """
    error: float = np.inf
    iter: int = 0

    while error >= tol:
        kabsch = Kabsch(verbose=verbose)
        mobile = np.asarray([kabsch.fit_transform(_, reference) for _ in mobile])
        new_reference: NDArray[(Any, ...), Float] = mobile.mean(axis=0)
        error = rmse(new_reference, reference)
        reference = new_reference.copy()

        iter += 1
        if verbose:
            logger.info("Iteration %d -- error: %.6f", iter, error)

    if not verbose:
        logger.info("Total iterations: %d -- error: %.6f", iter, error)

    return mobile



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Align trajectories to their average structure."""
import logging.config
import time
from pathlib import Path
from typing import Any
from typing import List

import click
import mdtraj as md
import numpy as np
from mdtraj.utils import in_units_of
from nptyping import Float
from nptyping import NDArray

from .. import _MASK
from .. import create_logging_dict
from .. import PathLike
from ..libs.align import align_trajectory
from ..libs.utils import get_average_structure
from ..libs.utils import get_positions


@click.command("align", short_help="Align trajectory to a reference")
@click.option(
    "-s",
    "--top",
    "topology",
    metavar="FILE",
    default=Path.cwd().joinpath("input.top").as_posix(),
    show_default=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Topology",
)
@click.option(
    "-f",
    "--traj",
    "trajectory",
    metavar="FILE",
    default=[
        Path.cwd().joinpath("input.nc").as_posix(),
    ],
    show_default=True,
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Trajectory",
)
@click.option(
    "-r",
    "--ref",
    "reference",
    metavar="FILE",
    default=Path.cwd().joinpath("average.pdb").as_posix(),
    show_default=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Average structure of trajectory",
)
@click.option(
    "-o",
    "--outfile",
    metavar="FILE",
    default=Path.cwd().joinpath("aligned.nc").as_posix(),
    show_default=True,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Aligned trajectory",
)
@click.option(
    "-l",
    "--logfile",
    metavar="LOG",
    show_default=True,
    default=Path.cwd().joinpath("align_traj.log").as_posix(),
    type=click.Path(exists=False, file_okay=True, dir_okay=False, resolve_path=True),
    help="Log file",
)
@click.option(
    "--dt",
    "step",
    metavar="OFFSET",
    default=0,
    show_default=True,
    type=click.IntRange(min=0, clamp=True),
    help="Trajectory output offset (0 = single step)",
)
@click.option(
    "-m",
    "--mask",
    default="ca",
    show_default=True,
    type=click.Choice(_MASK.keys()),
    help="Atom selection",
)
@click.option(
    "--tol",
    type=click.FloatRange(min=0.0, max=1.0),
    default=0.001,
    show_default=True,
    help="Error tolerance",
)
@click.option("-v", "--verbose", is_flag=True, help="Noisy output")
def cli(
    topology: PathLike,
    trajectory: List[str],
    reference: PathLike,
    outfile: PathLike,
    logfile: PathLike,
    step: int,
    mask: str,
    tol: float,
    verbose: bool,
) -> None:
    """Align a trajectory to average structure using Kabsch fitting."""
    start_time: float = time.perf_counter()

    # Setup logging
    logging.config.dictConfig(create_logging_dict(logfile))
    logger: logging.Logger = logging.getLogger(__name__)

    step = step if step > 0 else 1

    logger.info("Loading %s and %s", topology, trajectory)
    positions: NDArray[(Any, ...), Float] = get_positions(
        topology, trajectory, mask=_MASK[mask], stride=step
    )

    # Calculate average structure
    ref_traj: md.Trajectory = get_average_structure(
        topology, trajectory, mask=_MASK[mask], stride=step
    )

    logger.info("Saving average structure to %s", reference)
    ref_traj.save(reference)
    unitcell_angles: NDArray[(Any, ...), Float] = ref_traj.unitcell_angles.copy()
    unitcell_lengths: NDArray[(Any, ...), Float] = ref_traj.unitcell_lengths.copy()
    unitcell_vectors: NDArray[(Any, ...), Float] = ref_traj.unitcell_vectors.copy()
    if not (
        ".gro" in "".join(trajectory)
        or ".xtc" in "".join(trajectory)
        or ".trj" in "".join(trajectory)
        or ".tng" in "".join(trajectory)
    ):
        in_units_of(ref_traj.xyz, "nanometer", "angstroms", inplace=True)

    logger.info("Aligning trajectory to average structures")
    ref_traj.xyz = align_trajectory(
        positions, ref_traj.xyz[0], tol=tol, verbose=verbose
    )
    n_frames = ref_traj.n_frames
    ref_traj.time = np.arange(n_frames)
    ref_traj.unitcell_angles = np.repeat(unitcell_angles, n_frames, axis=0)
    ref_traj.unitcell_lengths = np.repeat(unitcell_lengths, n_frames, axis=0)
    ref_traj.unitcell_vectors = np.repeat(unitcell_vectors, n_frames, axis=0)
    if not (
        ".gro" in "".join(trajectory)
        or ".xtc" in "".join(trajectory)
        or ".trj" in "".join(trajectory)
        or ".tng" in "".join(trajectory)
    ):
        in_units_of(ref_traj.xyz, "angstroms", "nanometer", inplace=True)

    logger.info("Saving aligned trajectory to %s}", outfile)
    ref_traj.save(outfile)

    stop_time: float = time.perf_counter()
    dt: float = stop_time - start_time
    struct_time: time.struct_time = time.gmtime(dt)
    if verbose:
        output: str = time.strftime("%H:%M:%S", struct_time)
        logger.info(f"Total execution time: {output}")



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Module to find signals using the Jade ICA method.

This module contains the function, _jade, which does blind source separation of real
signals. The original Python code can be found at
https://github.com/gvacaliuc/jade_c/blob/master/jade.py
"""
from __future__ import annotations

import logging
from itertools import combinations
from typing import Any
from typing import Optional

import numpy as np
from nptyping import Float
from nptyping import NDArray
from scipy import linalg
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_array
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


def _jade(
    arr: NDArray[(Any, ...), Float], n_components: int = 1
) -> NDArray[(Any, ...), Float]:
    """Blind separation of real signals with JADE.

    jadeR implements JADE, an Independent Component Analysis (ICA) algorithm developed
    by Jean-Francois Cardoso. See http://www.tsi.enst.fr/~cardoso/guidesepsou.html, and
    papers cited at the end of the source file.

    Translated into NumPy from the original Matlab Version 1.8 (May 2005) by Gabriel
    Beckers, http://gbeckers.nl .

    Parameters
    ----------
    arr : NDArray
        a data matrix (n_features, n_samples)
    n_components : int
        output matrix B has size mxn so that only m sources are extracted.  This is done
        by restricting the operation of jadeR to the m first principal components.
        Defaults to None, in which case :math:`m = n`.

    Returns
    -------
    NDArray
        An m*n matrix B (NumPy matrix type), such that Y=B*X are separated sources
        extracted from the n*T data matrix X. If m is omitted, B is a square n*n matrix
        (as many sources as sensors). The rows of B are ordered such that the columns of
        :math:`pinv(B)` are in order of decreasing norm; this has the effect that the
        `most energetically significant` components appear first in the rows of
        :math:`Y = B * X`.

    Notes
    -----
    Quick notes (more at the end of this file):

    - This code is for REAL-valued signals.  A MATLAB implementation of JADE for both
      real and complex signals is also available from
      http://sig.enst.fr/~cardoso/stuff.html

    - This algorithm differs from the first released implementations of JADE in that it
      has been optimized to deal more efficiently
        1) with real signals (as opposed to complex)
        2) with the case when the ICA model does not necessarily hold.

    - There is a practical limit to the number of independent components that can be
      extracted with this implementation.  Note that the first step of JADE amounts to a
      PCA with dimensionality reduction from `n` to `m` (which defaults to n).  In
      practice m cannot be *very large* (more than 40, 50, 60... depending on available
      memory)

    - See more notes, references and revision history at the end of this file and more
      stuff at http://sig.enst.fr/~cardoso/stuff.html

    - For more info on NumPy translation, see the end of this file.

    - This code is supposed to do a good job!  Please report any problem relating to
      the NumPY code gabriel@gbeckers.nl

    Copyright original Matlab code : Jean-Francois Cardoso <cardoso@sig.enst.fr>
    Copyright Numpy translation : Gabriel Beckers <gabriel@gbeckers.nl>
    """
    logger.info("jade -> Looking for %d sources", n_components)
    logger.info("jade -> Removing the mean value")
    n_samples, _ = arr.shape
    arr -= arr.mean(axis=0)

    # whitening & projection onto signal subspace
    # ===========================================
    logger.info("jade -> Whitening the data")

    # --- PCA  ----------------------------------------------------------
    u, s, vh = linalg.svd(arr, full_matrices=False)
    u, s, vh = u[:, :n_components], s[:n_components], vh[:n_components]
    B = linalg.inv(np.diag(s)) @ vh * np.sqrt(n_samples)
    arr = u * np.sqrt(n_samples)

    del u, s, vh
    # NOTE: At this stage, X is a PCA analysis in m components of the real data, except
    # that all its entries now have unit variance.  Any further rotation of X will
    # preserve the property that X is a vector of uncorrelated components.  It remains
    # to find the rotation matrix such that the entries of X are not only uncorrelated
    # but also `as independent as possible".  This independence is measured by
    # correlations of order higher than 2.  We have defined such a measure of
    # independence which
    #   1) is a reasonable approximation of the mutual information
    #   2) can be optimized by a `fast algorithm"
    # This measure of independence also corresponds to the `diagonality" of a set of
    # cumulant matrices.  The code below finds the `missing rotation " as the matrix
    # which best diagonalizes a particular set of cumulant matrices.

    # Estimation of the cumulant matrices.
    # ====================================
    logger.info("jade -> Estimating cumulant matrices")

    # Dim. of the space of real symm matrices
    dimsymm = n_components * (n_components + 1) // 2
    nbcm = dimsymm  # number of cumulant matrices
    # Storage for cumulant matrices
    CM = np.zeros((n_components, n_components * nbcm))
    R = np.eye(n_components)

    # I am using a symmetry trick to save storage.  I should write a short note one of
    # these days explaining what is going on here.
    # will index the columns of CM where to store the cum. mats.
    Range = np.arange(n_components)
    sqrt2 = np.sqrt(2)

    for im in range(n_components):
        Xim = np.c_[arr[:, im]]
        Xijm = Xim ** 2
        # Note to myself: the -R on next line can be removed: it does not affect
        # the joint diagonalization criterion
        Rim = np.c_[R[:, im]]
        CM[:, Range] = (Xijm * arr).T @ arr / n_samples - R - 2 * (Rim @ Rim.T)
        Range += n_components
        for jm in range(im):
            Xijm = Xim * np.c_[arr[:, jm]]
            Rjm = np.c_[R[:, jm]]
            CM[:, Range] = (
                sqrt2 * (Xijm * arr).T @ arr / n_samples - Rim @ Rjm.T - Rjm @ Rim.T
            )
            Range = Range + n_components

    # Now we have nbcm = m(m+1)/2 cumulants matrices stored in a big m x m*nbcm array.

    V = np.eye(n_components)

    On = np.zeros(CM.shape[0])
    Range = np.arange(n_components)
    for _ in range(nbcm):
        diag = np.diag(CM[:, Range])
        On += np.sum(diag ** 2, axis=0)
        Range += n_components
    Off = np.sum(CM ** 2) - On

    # % A statistically scaled threshold on `small" angles
    seuil = 1e-6 / np.sqrt(n_samples)
    encore = True
    sweep = 0  # % sweep number
    updates = 0  # % Total number of rotations

    # Joint diagonalization proper
    logger.info("jade -> Contrast optimization by joint diagonalization")

    counters = list(combinations(range(n_components), 2))
    while encore:
        encore = False
        logger.info("jade -> Sweep #%3d", sweep)
        sweep += 1
        upds = 0  # % Number of rotations in a given seep

        for p, q in counters:
            Ip = np.arange(p, n_components * nbcm, n_components)
            Iq = np.arange(q, n_components * nbcm, n_components)

            # computation of Givens angle
            g = np.c_[[CM[p, Ip] - CM[q, Iq], CM[p, Iq] + CM[q, Ip]]]
            gg = g @ g.T
            ton = gg[0, 0] - gg[1, 1]
            toff = gg[0, 1] + gg[1, 0]
            theta = 0.5 * np.arctan2(toff, ton + np.sqrt(ton * ton + toff * toff))
            Gain = (np.sqrt(ton * ton + toff * toff) - ton) / 4.0

            # Givens update
            if abs(theta) > seuil:
                encore = True
                upds += 1
                c = np.cos(theta)
                s = np.sin(theta)
                G = np.array([[c, -s], [s, c]])
                pair = np.array([p, q])
                V[:, pair] = V[:, pair] @ G
                CM[pair, :] = G.T @ CM[pair, :]
                cIp = np.c_[CM[:, Ip]]
                cIq = np.c_[CM[:, Iq]]
                CM[:, np.r_[Ip, Iq]] = np.c_[c * cIp + s * cIq, -s * cIp + c * cIq]
                On += Gain
                Off -= Gain

        logger.info("completed in %d rotations", upds)
        updates = updates + upds
    logger.info("jade -> Total of %d Givens rotations", updates)

    # A separating matrix
    # ===================
    B = V.T @ B

    # Permute the rows of the separating matrix B to get the most energetic components
    # first. Here the **signals** are normalized to unit variance.  Therefore, the sort
    # is according to the norm of the columns of A = pinv(B)

    logger.info("jade -> Sorting the components")

    A = linalg.pinv(B)
    keys = np.argsort((A * A).sum(axis=0))[::-1]
    B = B[keys, :]

    logger.info("jade -> Fixing the signs")
    b = B[:, 0]
    signs = np.sign(np.sign(b) + 0.1)  # just a trick to deal with sign=0
    B = np.diag(signs) @ B

    return B

    # Revision history of MATLAB code:
    #
    # - V1.8, May 2005
    #  - Added some commented code to explain the cumulant computation tricks.
    #  - Added reference to the Neural Comp. paper.
    #
    # -  V1.7, Nov. 16, 2002
    #   - Reverted the mean removal code to an earlier version (not using
    #     repmat) to keep the code octave-compatible.  Now less efficient,
    #     but does not make any significant difference wrt the total
    #     computing cost.
    #   - Remove some cruft (some debugging figures were created.  What
    #     was this stuff doing there???)
    #
    #
    # -  V1.6, Feb. 24, 1997
    #   - Mean removal is better implemented.
    #   - Transposing X before computing the cumulants: small speed-up
    #   - Still more comments to emphasize the relationship to PCA
    #
    # -  V1.5, Dec. 24 1997
    #   - The sign of each row of B is determined by letting the first element be
    #     positive.
    #
    # -  V1.4, Dec. 23 1997
    #   - Minor clean up.
    #   - Added a verbose switch
    #   - Added the sorting of the rows of B in order to fix in some reasonable way the
    #     permutation indetermination.  See note 2) below.
    #
    # -  V1.3, Nov.  2 1997
    #   - Some clean up.  Released in the public domain.
    #
    # -  V1.2, Oct.  5 1997
    #   - Changed random picking of the cumulant matrix used for initialization to a
    #     deterministic choice.  This is not because of a better rationale but to make
    #     the ouput (almost surely) deterministic.
    #   - Rewrote the joint diag. to take more advantage of Matlab"s tricks.
    #   - Created more dummy variables to combat Matlab"s loose memory management.
    #
    # -  V1.1, Oct. 29 1997.
    #    Made the estimation of the cumulant matrices more regular. This also corrects a
    #    buglet...
    #
    # -  V1.0, Sept. 9 1997. Created.
    #
    # Main references:
    # @article{CS-iee-94,
    #  title 	= "Blind beamforming for non {G}aussian signals",
    #  author       = "Jean-Fran\c{c}ois Cardoso and Antoine Souloumiac",
    #  HTML 	= "ftp://sig.enst.fr/pub/jfc/Papers/iee.ps.gz",
    #  journal      = "IEE Proceedings-F",
    #  month = dec, number = 6, pages = {362-370}, volume = 140, year = 1993}
    #
    #
    # @article{JADE:NC,
    #  author  = "Jean-Fran\c{c}ois Cardoso",
    #  journal = "Neural Computation",
    #  title   = "High-order contrasts for independent component analysis",
    #  HTML    = "http://www.tsi.enst.fr/~cardoso/Papers.PS/neuralcomp_2ppf.ps",
    #  year    = 1999, month =	jan,  volume =	 11,  number =	 1,  pages =  "157-192"}
    #
    #
    #  Notes:
    #  ======
    #
    #  Note 1) The original Jade algorithm/code deals with complex signals in Gaussian
    #  noise white and exploits an underlying assumption that the model of independent
    #  components actually holds.  This is a reasonable assumption when dealing with
    #  some narrowband signals.  In this context, one may i) seriously consider dealing
    #  precisely with the noise in the whitening process and ii) expect to use the small
    #  number of significant eigenmatrices to efficiently summarize all the 4th-order
    #  information.  All this is done in the JADE algorithm.
    #
    #  In *this* implementation, we deal with real-valued signals and we do NOT expect
    #  the ICA model to hold exactly.  Therefore, it is pointless to try to deal
    #  precisely with the additive noise and it is very unlikely that the cumulant
    #  tensor can be accurately summarized by its first n eigen-matrices.  Therefore,
    #  we consider the joint diagonalization of the *whole* set of eigen-matrices.
    #  However, in such a case, it is not necessary to compute the eigenmatrices at all
    #  because one may equivalently use `parallel slices" of the cumulant tensor.  This
    #  part (computing the eigen-matrices) of the computation can be saved: it suffices
    #  to jointly diagonalize a set of cumulant matrices.  Also, since we are dealing
    #  with reals signals, it becomes easier to exploit the symmetries of the cumulants
    #  to further reduce the number of matrices to be diagonalized. These
    #  considerations, together with other cheap tricks lead to this version of JADE
    #  which is optimized (again) to deal with real mixtures and to work `outside the
    #  model'.  As the original JADE algorithm, it works by minimizing a `good set' of
    #  cumulants.
    #
    #  Note 2) The rows of the separating matrix B are resorted in such a way that the
    #  columns of the corresponding mixing matrix A=pinv(B) are in decreasing order of
    #  (Euclidian) norm.  This is a simple, `almost canonical" way of fixing the
    #  indetermination of permutation.  It has the effect that the first rows of the
    #  recovered signals (ie the first rows of B*X) correspond to the most energetic
    #  *components*.  Recall however that the source signals in S=B*X have unit
    #  variance.  Therefore, when we say that the observations are unmixed in order of
    #  decreasing energy, this energetic signature is to be found as the norm of the
    #  columns of A=pinv(B) and not as the variances of the separated source signals.
    #
    #  Note 3) In experiments where JADE is run as B=jadeR(X,m) with m varying in range
    #  of values, it is nice to be able to test the stability of the decomposition.  In
    #  order to help in such a test, the rows of B can be sorted as described above. We
    #  have also decided to fix the sign of each row in some arbitrary but fixed way.
    #  The convention is that the first element of each row of B is positive.
    #
    #  Note 4) Contrary to many other ICA algorithms, JADE (or least this version) does
    #  not operate on the data themselves but on a statistic (the full set of 4th order
    #  cumulant). This is represented by the matrix CM below, whose size grows as m^2 x
    #  m^2 where m is the number of sources to be extracted (m could be much smaller
    #  than n).  As a consequence, (this version of) JADE will probably choke on a
    #  `large' number of sources. Here `large' depends mainly on the available memory
    #  and could be something like 40 or so.  One of these days, I will prepare a
    #  version of JADE taking the `data' option rather than the `statistic' option.

    # Notes on translation (GB):
    # =========================
    #
    # Note 1) The function jadeR is a relatively literal translation from the original
    # MATLABcode. I haven't really looked into optimizing it for NumPy. If you have any
    # time to look at this and good ideas, let me know.
    #
    # Note 2) A test module that compares NumPy output with Octave (MATLAB
    # clone) output of the original MATLAB script is available


class JadeICA(TransformerMixin, BaseEstimator):
    """Perform blind source separation using joint diagonalization."""

    def __init__(self, *, n_components: Optional[int] = None) -> None:
        """Perform a blind signal separation using joint diagonalization.

        Parameters
        ----------
        n_components : int, default=None
            Number of signals to extract. `None` assumes all components
        """
        super().__init__()

        self.n_components: Optional[int] = n_components
        self.mean_: NDArray[(Any, ...), Float]
        self.components_: NDArray[(Any, ...), Float]

    def fit(
        self,
        arr: NDArray[(Any, ...), Float],
        y: Optional[NDArray[(Any, ...), Float]] = None,
    ) -> JadeICA:
        """Calculate the unmixing matrix.

        Parameters
        ----------
        arr : NDArray
            mixed signal array
        y : NDArray, optional
            unused

        Returns
        -------
        self

        Raises
        ------
        IndexError
            if :math:`m > n_features`
        """
        # GB: we do some checking of the input arguments and copy data to new variables to
        # avoid messing with the original input. We also require double precision (float64)
        # and a numpy matrix type for X.
        origtype = arr.dtype  # remember to return matrix B of the same type
        arr = check_array(arr, dtype=float)

        # GB: n is number of input signals, T is number of samples
        _, n_features = arr.shape

        if self.n_components is None:
            self.n_components = n_features  # Number of sources defaults to # of sensors
        elif self.n_components > n_features:
            raise IndexError(
                f"More sources ({self.n_components}) than sensors ({n_features})"
            )

        self.mean_ = arr.mean(axis=0)
        self.components_ = _jade(arr, n_components=self.n_components).astype(origtype)
        return self

    def transform(self, arr: NDArray[(Any, ...), Float]) -> NDArray[(Any, ...), Float]:
        """Project the unmixing matrix onto `arr` to give independent signals.

        Parameters
        ----------
        arr : NDArray
            Unmixed signals

        Returns
        -------
        NDArray
            Unmixed signal array
        """
        check_is_fitted(self)

        arr -= self.mean_
        signal: NDArray[(Any, ...), Float] = arr @ self.components_.T
        return signal

    def fit_transform(
        self,
        arr: NDArray[(Any, ...), Float],
        y: Optional[NDArray[(Any, ...), Float]] = None,
        **fit_params: str,
    ) -> NDArray[(Any, ...), Float]:
        """Determine the independent signals using joint diagonalization.

        Parameters
        ----------
        arr : NDArray
            Mixed signal array
        y : NDArray, optional
            unused
        fit_params : dict
            unused

        Returns
        -------
        NDArray
            Unmixed signal array

        Raises
        ------
        IndexError
            if :math:`m > n_features`
        """
        # GB: we do some checking of the input arguments and copy data to new variables to
        # avoid messing with the original input. We also require double precision (float64)
        # and a numpy matrix type for X.
        origtype = arr.dtype  # remember to return matrix B of the same type
        arr1 = check_array(arr, dtype=float)

        # GB: n is number of input signals, T is number of samples
        _, n_features = arr.shape

        if self.n_components is None:
            self.n_components = n_features  # Number of sources defaults to # of sensors
        elif self.n_components > n_features:
            raise IndexError(
                f"More sources ({self.n_components}) than sensors ({n_features})"
            )

        self.mean_ = arr.mean(axis=0)
        self.components_ = _jade(arr1, n_components=self.n_components).astype(origtype)

        arr -= self.mean_
        signal: NDArray[(Any, ...), Float] = arr @ self.components_.T
        return signal

    def inverse_transform(
        self, arr: NDArray[(Any, ...), Float]
    ) -> NDArray[(Any, ...), Float]:
        """Find the original mixed signals.

        Parameters
        ----------
        arr : NDArray
            Unmixed signal array

        Raises
        ------
        NotImplementedError
            If called because Jade does not provide inversion
        """
        raise NotImplementedError(
            "Inverse transformation is currently not implemented."
        )



# --------------------------------------------------------------------------------------
#  Copyright (C) 2020–2021 by Timothy H. Click <tclick@okstate.edu>
#
#  Permission to use, copy, modify, and/or distribute this software for any purpose
#  with or without fee is hereby granted.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
#  FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
#  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
#  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
#  THIS SOFTWARE.
# --------------------------------------------------------------------------------------
"""Decomposition package."""



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['qaa', 'qaa.commands', 'qaa.decomposition', 'qaa.libs']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'holoviews[recommended]>=1.14.3,<2.0.0',
 'mdtraj>=1.9,<2.0',
 'nptyping>=1.4,<2.0',
 'scikit-learn>=0.24,<0.25']

extras_require = \
{'jupyter': ['jupyter>=1.0.0,<2.0.0',
             'jupyterlab>=3.0.13,<4.0.0',
             'jupyterlab-code-formatter>=1.4.10,<2.0.0',
             'jupyterlab-mathjax3>=4.2.2,<5.0.0',
             'nglview>=3.0.0,<4.0.0',
             'plotly>=4.14.3,<5.0.0']}

entry_points = \
{'console_scripts': ['qaa = qaa.__main__:main']}

setup_kwargs = {
    'name': 'qaa',
    'version': '3.0.0rc3',
    'description': 'Quasi-Anharmonic Analysis',
    'long_description': 'Quasi-Anharmonic Analysis\n=========================\n\n|PyPI| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/qaa.svg\n   :target: https://pypi.org/project/qaa/\n   :alt: PyPI\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/qaa\n   :target: https://pypi.org/project/qaa\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/qaa\n   :target: https://opensource.org/licenses/BSD-3-Clause\n   :alt: License\n.. |Read the Docs| image:: https://readthedocs.org/projects/pyqaa/badge/?version=latest\n   :target: https://pyqaa.readthedocs.io/en/latest/?badge=latest\n   :alt: ReDocumentation Status\n.. |Tests| image:: https://github.com/tclick/qaa/workflows/Tests/badge.svg\n   :target: https://github.com/tclick/qaa/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/tclick/qaa/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/tclick/qaa\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. image:: https://pyup.io/repos/github/tclick/qaa/shield.svg\n     :target: https://pyup.io/repos/github/tclick/qaa/\n     :alt: Updates\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\nFeatures\n--------\n\n`qaa` analyzes molecular dynamics (MD) trajectories by using joint\ndiagonalization (JADE) to separate the information. The JADE [1]_ and QAA [2]_\ncode are based on the original code written in Matlab.\n\n.. [1] Cardoso, J. F.; Souloumiac, A. "Blind Beamforming for Non-Gaussian\n       Signals." IEE Proc F Radar Signal Process 1993, 140 (6), 362.\n.. [2] Ramanathan, A.; Savol, A. J.; Langmead, C. J.; Agarwal, P. K.;\n       Chennubhotla, C. S. "Discovering Conformational Sub-States Relevant to Protein\n       Function." Plos One 2011, 6 (1), e15827.\n\nRequirements\n------------\n\n* Python 3.8+\n* click 7.0+\n* numpy 1.20+\n* scipy 1.6+\n* matplotlib 3.3+\n* scikit-learn 0.24+\n* mdtraj 1.9+\n* nptyping 1.4+\n* holoviews 1.14+\n\nInstallation\n------------\n\nYou can install *Quasi-Anharmonic Analysis* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install qaa\n\nIf you want to visualize the tutorial notebooks, you can install the extra\ndependencies via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install qaa[jupyter]\n\n\nUsage\n-----\n\nPlease see the `Command-line Reference <Usage_>`_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `BSD 3 Clause license`_,\n*Quasi-Anharmonic Analysis* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _BSD 3 Clause license: https://opensource.org/licenses/BSD-3-Clause\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/tclick/qaa/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://qaa.readthedocs.io/en/latest/usage.html\n',
    'author': 'Timothy H. Click',
    'author_email': 'tclick@okstate.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tclick/qaa',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7, !=2.7.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',
}


setup(**setup_kwargs)
